from datetime import date
from enum import Enum

from hgraph import TimeSeriesSchema, TSS, subscription_service, TS, default_path, TSB, operator, reference_service, \
    compute_node, contains_, graph, TIME_SERIES_TYPE, last_modified_date, sample, if_true, not_

__all__ = ["HolidayCalendarSchema", "calendar_for", "Periods", "business_days", "business_day", "HolidayCalendar",
           "filter_by_calendar"]


class HolidayCalendarSchema(TimeSeriesSchema):
    """
    The typical schema here is to track holidays and the start and end of the week. With the days between
    start of week and end of week being weekends or non-working days.
    If this is not appropriate the logic should also be able to cope with no-weekends specified (i.e. sow and eow being
    monday and sunday respectively) and any actual weekends being encoded in the holidays. This supports a more flexible
    specification that supports weekends shifting (as is the case with the UAE recently).
    Alternatively the calendar could be adjusted using the point-in-time of the engine clock.
    This is an implementation choice and needs to be clearly specified so the user can make an appropriate decision.
    """
    holidays: TSS[date]
    start_of_week: TS[int]
    end_of_week: TS[int]


HolidayCalendar = TSB[HolidayCalendarSchema]


@subscription_service
def calendar_for(symbol: TS[str], path: str = default_path) -> HolidayCalendar:
    """
    The calendar service for a given symbol.
    """


class Periods(Enum):
    Week = 1
    Month = 2
    Quarter = 3
    Year = 4


@operator
def business_days(period: TS[Periods], calendar: HolidayCalendar, dt: TS[date]) -> TS[tuple[date, ...]]:
    """
    Identifies the business days for the given period, using the given calendar.
    This will be for the period containing the current engine clock or (if provided) the
    dt provided.
    """


@reference_service
def trade_date(path: str = default_path) -> TS[date]:
    """
    The current trade date for this process. The trade date is wired in to the outer graph and defines the sessions
    trading date. Whilst it is possible to have multiple trading dates simultaneously, this not the general case.
    When multiple trade dates are possible, the path would disambiguate the trading dates. Note that a trading date
    does not imply that a contract being considered for trading would be tradable, just that if it were to trade we would
    consider the trade date to be the date of record for the trade.
    """


@subscription_service
def business_day(symbol: TS[str], path: str = default_path) -> TS[date]:
    """
    This will make use of the trade_date to track if the current trade_date is also a business day for the instrument.
    This is slightly different to just walking the business days of a calendar, as if we are not trading and the instrument
    is tradeable, we will still not produce a date.
    """


@compute_node(overloads=contains_)
def _contains_dt_in_calendar(ts: HolidayCalendar, item: TS[date]) -> TS[bool]:
    """
    Determines if the date is within the holiday calendar, for us that means if we deem it a non-working day then
    the date is within the calendar.
    """
    dt = item.value
    dow = dt.weekday()
    sow = ts.start_of_week.value
    eow = ts.end_of_week.value
    if eow < sow:
        if eow < dow < sow:
            return True  # We are in a weekend
    else:
        if dow < sow or eow < dow:
            return True  # The other perspective for weekend
    return dt in ts.holidays.value


@graph
def filter_by_calendar(ts: TIME_SERIES_TYPE, holidays: HolidayCalendar) -> TIME_SERIES_TYPE:
    """Restrict values to be published only during working days"""
    dt = last_modified_date(ts)
    return sample(if_true(not_(contains_(holidays, dt))), ts)



from frozendict import frozendict
from hgraph import service_impl, generator, TSS, TSD, TS, TimeSeriesSchema, CompoundScalar, Frame, graph
from datetime import date, datetime

from hg_systematic.operators import price_in_dollars

__all__ = ["price_in_dollars_static_impl", "StaticPriceSchema"]


class StaticPriceSchema(CompoundScalar):
    """A very simplistic schema for replaying prices"""
    date: date
    symbol: str
    price: float


@service_impl(interfaces=[price_in_dollars])
@graph
def price_in_dollars_static_impl(symbol: TSS[str], prices: Frame[StaticPriceSchema]) -> TSD[str, TS[float]]:
    return _price_in_dollars_static_impl(prices)


@generator
def _price_in_dollars_static_impl(prices: Frame[StaticPriceSchema]) -> TSD[str, TS[float]]:
    # This approach (ticking everything without filter, etc. has the consequence of not having value immediately
    # available, however, in a dynamic implementation this would have a delay between the time of request to
    # value being available. This could have consequences when swapping out implementations.
    for dt, prices_ in prices.sort(by="date").cast({'date': datetime}).partition_by("date", maintain_order=True, include_key=False, as_dict=True).items():
        yield dt[0], frozendict(prices_.iter_rows())



import calendar as cal
from datetime import date, timedelta, datetime
from typing import Iterable

from frozendict import frozendict as fd
from hgraph import TS, compute_node, \
    graph, service_impl, default_path, contains_, if_true, sample, TSS, TSD, map_, not_, \
    EvaluationEngineApi, generator, const

from hg_systematic.operators._calendar import Periods, business_days, business_day, calendar_for, \
    trade_date, HolidayCalendar

__all__ = ["business_day_impl", "trade_date_week_days", "calendar_for_static", "create_market_holidays",
           "holiday_const"]


@compute_node(overloads=business_days)
def business_days_impl(period: TS[Periods], calendar: HolidayCalendar, dt: TS[date],
                       _output: TS[tuple[date, ...]] = None) -> TS[tuple[date, ...]]:
    dt = dt.value
    if not period.modified and not calendar.modified and _output.valid and len(dts := _output.value) > 1:
        # Check if the date is still within the bounds, if it is then no further work required
        if dts[0] <= dt <= dts[-1]:
            return  # Don't tick any change

    sow = calendar.start_of_week.value
    eow = calendar.end_of_week.value
    holidays = calendar.holidays.value
    if holidays is None:
        holidays = frozenset()
    period = period.value
    weekends = {(eow + d) % 7 for d in range(1, (sow - eow - 1) % 7 + 1)}
    if period == Periods.Week:
        dow = dt.weekday()
        start_dt = dt - timedelta(days=(dow - sow) if dow >= sow else (dow + 7 - sow))
        count = 7 - len(weekends)
    elif period == Periods.Month:
        start_dt = dt.replace(day=1)
        count = cal.monthrange(dt.year, dt.month)[1]
    elif period == Periods.Quarter:
        start_dt = dt.replace(month=(dt.month - 1) // 3 * 3 + 1, day=1)
        count = sum(cal.monthrange(start_dt.year, start_dt.month + i)[1] for i in range(3))
    elif period == Periods.Year:
        start_dt = dt.replace(month=1, day=1)
        count = 366 if (dt.year % 4 == 0 and dt.year % 100 != 0) or (dt.year % 400 == 0) else 365
    else:
        raise ValueError(f"Unknown period {period}")
    days = tuple(dt_ for d in range(count) if
                 (dt_ := start_dt + timedelta(days=d)) not in holidays and dt_.weekday() not in weekends)
    return days


@service_impl(interfaces=(business_day,))
def business_day_impl(symbol: TSS[str], calendar_path: str = default_path, trade_date_path: str = default_path) -> TSD[
    str, TS[date]]:
    return map_(_business_day_impl, calendar_path="" if calendar_path is None else calendar_path,
                trade_date_path="" if trade_date_path is None else trade_date_path, __keys__=symbol)


@graph
def _business_day_impl(key: TS[str], calendar_path: str, trade_date_path: str) -> TS[date]:
    calendar = calendar_for(key, path=default_path if calendar_path == "" else calendar_path)
    dt = trade_date(path=default_path if trade_date_path == "" else trade_date_path)
    return sample(if_true(not_(contains_(calendar.holidays, dt))), dt)


@service_impl(interfaces=(trade_date,))
@generator
def trade_date_week_days(sow: int = 0, eow: int = 4, _api: EvaluationEngineApi = None) -> TS[date]:
    """
    Provides a trade-date generator over all weekdays. If the trade date follows a non-traditional Sat, Sun weekend,
    then supply the appropriate start and end of week.
    """
    dt = _api.start_time.date()
    end_date = _api.end_time.date()
    while dt <= end_date:
        if sow <= dt.weekday() <= eow:
            yield datetime(dt.year, dt.month, dt.day), dt
        dt += timedelta(days=1)


@service_impl(interfaces=(calendar_for,))
def calendar_for_static(symbol: TSS[str], holidays: fd[str, frozenset[date]], sow: int = 0, eow: int = 4) -> TSD[
    str, HolidayCalendar]:
    """Provide a simple stub solution to provide holiday calendars from a fixed source of holidays."""
    holidays = const(holidays, tp=TSD[str, TSS[date]])
    return map_(
        lambda hols: HolidayCalendar.from_ts(holidays=hols, start_of_week=const(sow), end_of_week=const(eow)),
        holidays,
        __keys__=symbol
    )


def create_market_holidays(countries: Iterable[str], start_date_time: datetime, end_date_time: datetime) -> frozenset[
    date]:
    """Uses the holidays package to generate out holidays for the country codes supplied"""
    import holidays
    out = set()
    years = list(range(start_date_time.year, end_date_time.year + 1))
    for ctry in countries:
        out.update(holidays.country_holidays(ctry, years=years).keys())
    return frozenset(out)


def holiday_const(holidays: frozenset[date], sow: int = 0, eow: int = 4) -> HolidayCalendar:
    """Light-weight helper for testing with holidays"""
    return const(fd(holidays=holidays, start_of_week=sow, end_of_week=eow), tp=HolidayCalendar)



from hgraph import subscription_service, TS, default_path


__all__ = ["price_in_dollars", "SETTLEMENT_PRICE", "LIVE_PRICE", "returns_in_dollars",]

# Ticks with the settlement price when available.
SETTLEMENT_PRICE = "settlement_price"

# Ticks with the live price
LIVE_PRICE = "live_price"

@subscription_service
def price_in_dollars(symbol: TS[str], path: str = default_path) -> TS[float]:
    """
    Represent the current price of this symbol in dollars. The symbol can represent both simple and complex things.
    """


@subscription_service
def returns_in_dollars(symbol: TS[str], path: str = default_path) -> TS[float]:
    """
    The current returns in dollars for this symbol. This is based on close-to-close prices.
    """

